-- Foreign Language Support routines for Euphoria editor

--              Syntax Color
-- Break Euphoria statements into words with multiple colors.
-- The editor and pretty printer (eprint.ex) both use this file.

-- The user must define the following symbols to be colors for the
-- various syntax classes:
--       NORMAL_COLOR
--      COMMENT_COLOR
--      KEYWORD_COLOR
--      BUILTIN_COLOR
--       STRING_COLOR
--      BRACKET_COLOR  (a sequence of colors)

-- Euphoria keywords and routines built in to ex.exe/exw.exe

global constant keywords = {
    "if", "end", "then", "procedure", "else", "for", "return",
    "do", "elsif", "while", "type", "constant", "to", "and", "or",
    "exit", "function", "global", "by", "not", "include",
    "with", "without", "xor"}

global constant keywords_fl = { --change these to any language
    "se", "fim", "ento", "procedimento", "seno", "para", "devolver",
    "faz", "ou_se", "enquanto", "tipo", "constante", "at", "_e_", "ou",
    "sair", "funo", "global", "por", "no", "incluir",
    "com", "sem", "xou"}

global constant builtins = {
    "length",   "puts",         "integer",      "sequence",     "position",
    "object",   "append",       "prepend",      "print",        "printf",
    "?",        "clear_screen", "floor",        "getc",         "gets",
    "get_key",  "rand",         "repeat",       "atom",         "compare",
    "find",     "match",        "time",         "command_line", "open",
    "close",    "trace",        "getenv",       "sqrt",         "sin",
    "cos",      "tan",          "log",          "system",       "date",
    "remainder","power",        "machine_func", "machine_proc", "abort",
    "peek",     "poke",         "call",         "sprintf",      "arctan",
    "and_bits", "or_bits",      "xor_bits",     "not_bits",     "pixel",
    "get_pixel","mem_copy",     "mem_set",      "c_proc",       "c_func",
    "routine_id","call_proc",   "call_func",    "poke4",        "peek4s",
    "peek4u",   "profile",      "equal",        "system_exec",  "platform"}

global constant builtins_fl = { --change these to any language
    "tamanho",   "escrever",    "inteiro",      "sequncia",    "posio",
    "objecto",   "anexar",      "prepor",       "imprimir_objecto","imprimir_forma",
    "?",         "limpar_ecr", "cho",         "ler_carac",     "ler_linha",
    "ler_tecla", "aleat",       "repetir",      "tomo",         "comparar",
    "encontrar", "encontrar_fata", "tempo",    "linha_do_comando","abrir",
    "fechar",    "trilhar",     "ler_ambiente", "raizq",         "seno",
    "cos",       "tan",         "log",          "sistema",       "data",
    "resto",     "poder",       "funo_m quina","proc_m quina", "terminar",
    "olhar",     "meter",       "chamar",       "imprimir_forma_frase","arctan",
    "e_bits",    "ou_bits",     "xou_bits",     "no_bits",      "pixel",
    "ler_pixel", "copiar_mem",  "meter_mem",     "c_proc",       "c_func",
    "rotina_id", "chamar_proc", "chamar_func",   "meter4",       "olhar4s",
    "olhar4u",   "perfil",      "igual",         "sistema_exec", "plataforma"}

global constant library = {
    "all_palette",    "allocate",      "allocate_low",    "allocate_string",
    "allow_break",    "arccos",        "arcsin",          "atom_to_float32",
    "atom_to_float64","bits_to_int",   "bk_color",        "bytes_to_int",
    "call_back",      "chdir",         "check_break",     "crash_file",
    "crash_message",  "current_dir",   "cursor",          "custom_sort",
    "define_c_func",  "define_c_proc", "define_c_var",    "dir",
    "display_image","display_text_image","dos_interrupt", "draw_line",
    "ellipse",        "float32_to_atom","float64_to_atom","flush",
    "free",           "free_console",  "free_low",        "get",
    "get_active_page","get_all_palette","get_bytes",      "get_display_page",
    "get_mouse",      "get_position",   "get_screen_char","get_vector",
    "graphics_mode",  "instance",       "int_to_bits",    "int_to_bytes",
    "lock_file",      "lock_memory",    "lower",          "message_box",
    "mouse_events",   "mouse_pointer",  "open_dll",       "palette",
    "PI",             "polygon",        "prompt_number",  "prompt_string",    -- 2.1
    "put_screen_char","read_bitmap",    "register_block", "reverse",          -- 2.1
    "save_bitmap",    "save_image",     "save_screen",    "save_text_image",
    "scroll",         "seek",           "set_active_page","set_display_page",
    "set_rand",       "set_vector",     "sleep",          "sort",
    "sound",          "sprint",         "text_color",     "text_rows",
    "tick_rate",      "unlock_file",   "unregister_block","upper",
    "use_vesa",       "value",          "video_config",   "wait_key",
    "walk_dir",       "where",          "wildcard_file",  "wildcard_match",
    "wrap",
    -- The standard library files
    "dll.e",          "file.e",         "get.e",          "graphics.e",
    "image.e",        "machine.e",      "misc.e",         "mouse.e",
    "msgbox.e",       "safe.e",         "sort.e",         "wildcard.e"}

global constant library_fl = { --change these to any language
    "toda_palete",    "alocar",         "alocar_baixo",   "alocar_frase",
    "permite_terminar","arccos",        "arcsin",         " tom_...o_float32",
    " tom_...o_float64", "bits_...o_int",   "cor_fundo",      "bytes_...o_int",
    "volta_chamar",    "mudar_dir",     "ver_terminar",   "ficheiro_do_error",
    "mensagem_do_error","dir_actual",   "cursor",         "ordenar_opo",
    "definir_c_func",  "definir_c_proc","definir_c_var",  "dir",
    "mostrar_imagem",  "mostrar_imagem_texto","interrupo_dos","linha",
    "elipse",         "float32_...o_ tomo","float64_...o_ tomo","vazar",
    "libertar",       "libertar_janela", "libertar_baixo","ler_objecto",
    "obter_p gina_escrita","obter_palete","ler_bytes",    "obter_p gina_visual",
    "ler_rato",       "ler_posio",    "ler_carac_ecr", "obter_vector",
    "modo_gr fico",   "ocurncia",      "int_...o_bits",    "int_...o_bytes",
    "trancar_ficheiro","trancar_memoria","minscula",     "caixa_de_mensagem",
    "eventos_rato",   "ponteiro_rato",  "abrir_dll",      "palete",
    "PI",             "polgono",       "pedir_nmero",   "pedir_frase",
    "por_carac_ecr", "ler_bitmap",     "registo_bloco",  "contr rio",
    "guardar_bitmap", "guardar_imagem", "guardar_ecr",   "guardar_imagem_texto",
    "enrolar",        "ir_at",   "mudar_p gina_escrita", "mudar_p gina_visual",
    "valor_aleat",    "meter_vector",   "dormir",         "ordenar",
    "som",            "imprimir_frase", "cor_texto",      "linhas_texto",
    "velocidade", "distrancar_ficheiro","unregisto_bloco","maiscula",
    "utilizar_vesa",  "valor",          "config_vdeo",   "esperar_tecla",
    "todos_dir",      "onde",        "emparelhe_ficheiro","emparelhe_padro",
    "envoltura",
    -- The standard library files
    "dll.e",          "file.e",         "get.e",          "graphics.e",
    "image.e",        "machine.e",      "misc.e",         "mouse.e",
    "msgbox.e",       "safe.e",         "sort.e",         "wildcard.e"}

global constant with_clauses = {
  --"profile", -- Also a builtin
    "profile_time",
  --"trace",   -- Also a builtin
    "type_check",
    "warning"}--,
    --"protected_script"} -- Saffi

-- character classes
constant DIGIT = 1,
	 OTHER = 2,
	 LETTER  = 3,
	 BRACKET = 4,
	 QUOTE   = 5,
	 DASH = 6,
	 WHITE_SPACE = 7,
	 NEW_LINE = 8

sequence char_class

global procedure init_class()
-- set up character classes for easier line scanning
-- (assume no 0 char)
    char_class = repeat(OTHER, 255)

    char_class['a'..'z'] = LETTER
    char_class['A'..'Z'] = LETTER

    char_class[''..''] = LETTER   --change these for whatever language
    char_class[''..''] = LETTER
    char_class[''..''] = LETTER
    char_class[''..''] = LETTER
    char_class[''..''] = LETTER

    char_class['_'] = LETTER
    char_class['0'..'9'] = DIGIT
    char_class['['] = BRACKET
    char_class[']'] = BRACKET
    char_class['('] = BRACKET
    char_class[')'] = BRACKET
    char_class['{'] = BRACKET
    char_class['}'] = BRACKET
    char_class['\''] = QUOTE
    char_class['"'] = QUOTE
    char_class[' '] = WHITE_SPACE
    char_class['\t'] = WHITE_SPACE
    char_class['\r'] = WHITE_SPACE
    char_class['\n'] = NEW_LINE
    char_class['-'] = DASH
end procedure

constant DONT_CARE = -1  -- any color is ok - blanks, tabs

sequence line  -- the line being processed
integer seg_start, seg_end -- start and end of current segment of line
integer current_color
sequence color_segments -- the value returned
sequence english_line   -- the line returned

procedure seg_flush(integer new_color)
-- if the color must change,
-- add the current color segment to the sequence
-- and start a new segment
    if new_color != current_color then
	if current_color != DONT_CARE then
	    color_segments = append(color_segments,
	       {current_color, line[seg_start..seg_end]})
	    seg_start = seg_end + 1
	end if
	current_color = new_color
    end if
end procedure

global function SyntaxColor(sequence pline)
-- Break up a '\n'-terminated line into colored text segments identifying the
-- various parts of the Euphoria language.
-- Consecutive characters of the same color are all placed in the
-- same 'segment' - seg_start..seg_end.
-- A sequence is returned that looks like:
--     {{color1, "text1"}, {color2, "text2"}, ... }
    integer class, last, i, c, bracket_level
    sequence word

    line = pline
    current_color = DONT_CARE
    bracket_level = 0
    seg_start = 1
    seg_end = 0
    color_segments = {}

    while 1 do
	c = line[seg_end+1]
	class = char_class[c]

	if class = WHITE_SPACE then
	    seg_end += 1  -- continue with current color

	elsif class = LETTER then
	    last = length(line)-1
	    for j = seg_end + 2 to last do
		c = line[j]
		class = char_class[c]
		if class != LETTER then
		    if class != DIGIT then
			last = j - 1
			exit
		    end if
		end if
	    end for
	    word = line[seg_end+1..last]
	    if find(word, keywords_fl) then
		seg_flush(KEYWORD_COLOR)
	    elsif find(word, builtins_fl) then
		seg_flush(BUILTIN_COLOR)
	    elsif find(word, library_fl) then
		seg_flush(LIBRARY_COLOR)
	    else
		seg_flush(NORMAL_COLOR)
	    end if
	    seg_end = last

	elsif class <= OTHER then -- DIGIT too
	    seg_flush(NORMAL_COLOR)
	    seg_end += 1

	elsif class = BRACKET then
	    if find(c, "([{") then
		bracket_level += 1
	    end if
	    if bracket_level >= 1 and
	       bracket_level <= length(BRACKET_COLOR) then
		seg_flush(BRACKET_COLOR[bracket_level])
	    else
		seg_flush(NORMAL_COLOR)
	    end if
	    if find(c, ")]}") then
		bracket_level -= 1
	    end if
	    seg_end += 1

	elsif class = NEW_LINE then
	    exit  -- end of line

	elsif class = DASH then
	    if line[seg_end+2] = '-' then
		seg_flush(COMMENT_COLOR)
		seg_end = length(line)-1
		exit
	    end if
	    seg_flush(NORMAL_COLOR)
	    seg_end += 1

	else  -- QUOTE
	    i = seg_end + 2
	    while i < length(line) do
		if line[i] = c then
		    i += 1
		    exit
		elsif line[i] = '\\' then
		    if i < length(line)-1 then
			i += 1 -- ignore escaped char
		    end if
		end if
		i += 1
	    end while
	    seg_flush(STRING_COLOR)
	    seg_end = i - 1
	end if
    end while
    -- add the final piece:
    if current_color = DONT_CARE then
	current_color = NORMAL_COLOR
    end if
    return append(color_segments, {current_color, line[seg_start..seg_end]})
end function

global function Translate(sequence pline)
-- Break up a '\n'-terminated line into text segments identifying the
-- various parts of the Euphoria language.
-- A translated sequence is returned.

    integer class, last, i, c, bracket_level
    sequence word
    integer location location = 0
    line = pline
    bracket_level = 0
    seg_start = 1
    seg_end = 0
    english_line = {}

    while 1 do
	c = line[seg_end+1]
	class = char_class[c]

	if class = WHITE_SPACE then
	    seg_end += 1  -- continue
	    english_line &= c
	elsif class = LETTER then
	    last = length(line)-1
	    for j = seg_end + 2 to last do
		c = line[j]
		class = char_class[c]
		if class != LETTER then
		    if class != DIGIT then
			last = j - 1
			exit
		    end if
		end if
	    end for
	    word = line[seg_end+1..last]
	    location=find(word, keywords_fl)
	    if location then             --translate it
	      english_line &= keywords[location]
	    elsif find(word, builtins_fl) then
	      location=find(word, builtins_fl)
	      english_line &= builtins[location]
	    elsif find(word, library_fl) then
	      location=find(word, library_fl)
	      english_line &= library[location]
	    else
	      english_line &= word
	    end if
	    seg_end = last
	elsif class <= OTHER then -- DIGIT too
	    seg_end += 1
	    english_line &= c
	elsif class = BRACKET then
	    seg_end += 1
	    english_line &= c
	elsif class = NEW_LINE then
	    english_line &= c
	    exit  -- end of line

	elsif class = DASH then
	    seg_end += 1
	    english_line &= c
	else  -- QUOTE
	    i = seg_end + 2
	    seg_start=seg_end+1
	    while i < length(line) do
		if line[i] = c then  --found the matching quote mark
		    i += 1
		    exit
		elsif line[i] = '\\' then
		    if i < length(line)-1 then
			i += 1 -- ignore escaped char
		    end if
		end if
		i += 1
	    end while
	    seg_end = i - 1
	    english_line &= line[seg_start..seg_end]
	end if
    end while
    return english_line
end function


